@@ -4,9 +4,9 @@ import logging
 import pprint
 import random
 import uuid
-from typing import Awaitable, Callable, List, Optional
+from typing import Any, Awaitable, Callable, List, Optional, cast, Dict
 
-from aiohttp import ClientSession, WSMsgType, client_exceptions
+from aiohttp import ClientSession, ClientWebSocketResponse, WSMsgType, client_exceptions
 
 from .event import Event
 from .model.driver import Driver
@@ -54,9 +54,9 @@ class Client:
         self.aiohttp_session = aiohttp_session
         self.driver: Optional[Driver] = None
         # The WebSocket client
-        self.client = None
+        self.client: Optional[ClientWebSocketResponse] = None
         # Scheduled sleep task till next connection retry
-        self.retry_task = None
+        self.retry_task: Optional[asyncio.Task] = None
         # Boolean to indicate if we wanted the connection to close
         self.close_requested = False
         # The current number of attempts to connect, impacts wait time
@@ -67,9 +67,9 @@ class Client:
         self._on_disconnect: List[Callable[[], Awaitable[None]]] = []
         self._on_initialized: List[Callable[[], Awaitable[None]]] = []
         self._logger = logging.getLogger(__package__)
-        self._disconnect_event = None
+        self._disconnect_event: Optional[asyncio.Event] = None
 
-    def async_handle_message(self, msg) -> None:
+    def async_handle_message(self, msg: dict) -> None:
         """Handle incoming message.
 
         Run all async tasks in a wrapper to log appropriately.
@@ -77,7 +77,7 @@ class Client:
         # Right now only result is from receiving state.
         if msg["type"] == "result":
             if self.driver is None:
-                self.driver = Driver(msg["result"]["state"])
+                self.driver = Driver(self, msg["result"]["state"])
                 self._logger.info(
                     "Z-Wave JS initialized. %s nodes", len(self.driver.controller.nodes)
                 )
@@ -102,24 +102,53 @@ class Client:
         event = Event(type=msg["event"]["event"], data=msg["event"])
         self.driver.receive_event(event)
 
-    def register_on_connect(self, on_connect_cb: Callable[[], Awaitable[None]]):
+    def register_on_connect(
+        self, on_connect_cb: Callable[[], Awaitable[None]]
+    ) -> Callable:
         """Register an async on_connect callback."""
+
+        def unsubscribe() -> None:
+            """Unsubscribe listeners."""
+            if on_connect_cb in self._on_connect:
+                self._on_connect.remove(on_connect_cb)
+
         self._on_connect.append(on_connect_cb)
+        return unsubscribe
 
-    def register_on_disconnect(self, on_disconnect_cb: Callable[[], Awaitable[None]]):
+    def register_on_disconnect(
+        self, on_disconnect_cb: Callable[[], Awaitable[None]]
+    ) -> Callable:
         """Register an async on_disconnect callback."""
-        self._on_disconnect.append(on_disconnect_cb)
 
-    def register_on_initialized(self, on_initialized_cb: Callable[[], Awaitable[None]]):
+        def unsubscribe() -> None:
+            """Unsubscribe listeners."""
+            if on_disconnect_cb in self._on_disconnect:
+                self._on_disconnect.remove(on_disconnect_cb)
+
+        self._on_disconnect.append(on_disconnect_cb)
+        return unsubscribe
+
+    def register_on_initialized(
+        self, on_initialized_cb: Callable[[], Awaitable[None]]
+    ) -> Callable:
         """Register an async on_initialized_cb callback."""
+
+        def unsubscribe() -> None:
+            """Unsubscribe listeners."""
+            if on_initialized_cb in self._on_initialized:
+                self._on_initialized.remove(on_initialized_cb)
+
         self._on_initialized.append(on_initialized_cb)
+        return unsubscribe
 
     @property
-    def connected(self):
+    def connected(self) -> bool:
         """Return if we're currently connected."""
         return self.state == STATE_CONNECTED
 
-    async def async_send_json_message(self, message):
+    async def async_send_json_message(
+        self, message: Dict[str, Any], message_id: str = None
+    ) -> None:
         """Send a message.
 
         Raises NotConnected if client not connected.
@@ -130,9 +159,13 @@ class Client:
         if self._logger.isEnabledFor(logging.DEBUG):
             self._logger.debug("Publishing message:\n%s\n", pprint.pformat(message))
 
+        assert self.client
+        # include default messageId if needed
+        if message_id is None:
+            message["messageId"] = uuid.uuid4().hex
         await self.client.send_json(message)
 
-    async def connect(self):
+    async def connect(self) -> None:
         """Connect to the IoT broker."""
         if self.state != STATE_DISCONNECTED:
             raise RuntimeError("Connect called while not disconnected")
@@ -142,7 +175,7 @@ class Client:
         self.tries = 0
         self._disconnect_event = asyncio.Event()
 
-        while True:
+        while not self.close_requested:
             try:
                 self._logger.debug("Trying to connect")
                 await self._handle_connection()
@@ -151,15 +184,20 @@ class Client:
                 # Still adding it here to make sure we can always reconnect
                 self._logger.exception("Unexpected error")
 
-            if self.state == STATE_CONNECTED and self._on_disconnect:
-                await gather_callbacks(
-                    self._logger, "on_disconnect", self._on_disconnect
+            if self.state == STATE_CONNECTED:
+                # change state to connecting/disconnected
+                self.state = (
+                    STATE_DISCONNECTED if self.close_requested else STATE_CONNECTING
                 )
+                # notify callbacks about disconnection
+                if self._on_disconnect:
+                    await gather_callbacks(
+                        self._logger, "on_disconnect", self._on_disconnect
+                    )
 
             if self.close_requested:
                 break
 
-            self.state = STATE_CONNECTING
             self.tries += 1
 
             try:
@@ -172,7 +210,7 @@ class Client:
         self._disconnect_event.set()
         self._disconnect_event = None
 
-    async def _wait_retry(self):
+    async def _wait_retry(self) -> None:
         """Wait until it's time till the next retry."""
         # Sleep 2^tries + 0…tries*3 seconds between retries
         self.retry_task = asyncio.create_task(
@@ -181,9 +219,9 @@ class Client:
         await self.retry_task
         self.retry_task = None
 
-    async def _handle_connection(
+    async def _handle_connection(  # pylint: disable=too-many-branches, too-many-statements
         self,
-    ):  # pylint: disable=too-many-branches, too-many-statements
+    ) -> None:
         """Connect to the Z-Wave JS server."""
         client = None
         disconnect_warn = None
@@ -207,9 +245,7 @@ class Client:
             if self._on_connect:
                 await gather_callbacks(self._logger, "on_connect", self._on_connect)
 
-            await client.send_json(
-                {"messageId": uuid.uuid4().hex, "command": "start_listening"}
-            )
+            await self.async_send_json_message({"command": "start_listening"})
 
             while not client.closed:
                 msg = await client.receive()
@@ -234,8 +270,9 @@ class Client:
                 if self._logger.isEnabledFor(logging.DEBUG):
                     self._logger.debug("Received message:\n%s\n", pprint.pformat(msg))
 
+                msg_ = cast(dict, msg)
                 try:
-                    self.async_handle_message(msg)
+                    self.async_handle_message(msg_)
 
                 except InvalidState as err:
                     disconnect_warn = f"Invalid state: {err}"
@@ -261,7 +298,7 @@ class Client:
             else:
                 self._logger.warning("Connection closed: %s", disconnect_warn)
 
-    async def disconnect(self):
+    async def disconnect(self) -> None:
         """Disconnect the client."""
         self.close_requested = True
 
